import { TextEditor } from './text_editor';
import { ImageEditor } from './image_editor';
import AddText from './add_text.js';
import AddImage from './add_image.js';

export class ContentContainer {
  constructor(options) {
    this.pageViewer = options.pageViewer
    this.eventBus = options.eventBus
    this.pageIndex = options.pageIndex
    this.pagePtr = options.pagePtr
    this.messageHandler = options.messageHandler
    this.pageDiv = options.pageDiv
    this.viewport = options.viewport
    this.scale = options.scale
    this.$t = options.$t
    this.historyManager = options.contentEditHistoryManager

    this.destroyed = false
    this.rendered = false
    this.isRendering = false
    this.frameEditorList = []
    this.tool = this.pageViewer.tool || options.tool || ''
    this.color = this.pageViewer.color || ''
    this._selectedFrameIndex = -1
    this.initedFont = false

    this.onHandleTool = this.handleTool.bind(this)
    this.onKeydown = this.handleKeyDown.bind(this)
  }

  async init() {
    await this.update()

    document.addEventListener('keydown', this.onKeydown)
    this.eventBus._on('toolChanged', this.onHandleTool)
    this.render()
  }

  async update() {
    const { editPagePtr } = await this.messageHandler.sendWithPromise('InitEditPage', {
      pagePtr: this.pagePtr
    })
    this.editPagePtr = editPagePtr
    await this.initFont()

    const result = await this.messageHandler.sendWithPromise('BeginEdit', {
      editPagePtr: this.editPagePtr,
      type: 1 | 2
    })
    if (result && result.code) {
      console.log(result.message)
      alert(result.message)
      return false
    }

    this.tool = this.pageViewer.tool || ''
  }

  async initFont() {
    if (this.initedFont || this.tool === 'editImage' || this.tool === 'addImage') return
    await this.messageHandler.sendWithPromise('SetFontCallBackForEditPage', {
      editPagePtr: this.editPagePtr,
      fontFile: this.pageViewer.fontFile,
      fontName: 'DroidSansFallbackFull'
    })
    this.initedFont = true
  }

  async render () {
    if (this.contentContainer) {
      if (!this.initedFont) await this.initFont()
      return;
    }
    const rendering = this.isRendering
    this.isRendering = true

    const contentContainer = document.createElement('div')
    contentContainer.className = 'contentContainer'
    this.contentContainer = contentContainer
    this.pageDiv.append(this.contentContainer)

    this.frameEditorList.forEach(editor => {
      editor.removeViewer()
    })
    this.frameEditorList.length = 0

    try {
      const editAreaCount = await this.messageHandler.sendWithPromise('GetEditAreaCount', this.editPagePtr)

      for (let i = 0; i < editAreaCount; i++) {
        if (rendering) break

        const editAreaPtr = await this.messageHandler.sendWithPromise('GetEditArea', {
          editPagePtr: this.editPagePtr,
          index: i
        })
        if (!editAreaPtr) continue

        const isTextPtr = await this.messageHandler.sendWithPromise('IsTextArea', editAreaPtr)
        if (isTextPtr) {
          const frameEditor = new TextEditor({
            eventBus: this.eventBus,
            contentContainer: this,
            container: this.contentContainer,
            pagePtr: this.pagePtr,
            editPagePtr: this.editPagePtr,
            editAreaPtr,
            editAreaIndex: i,
            viewport: this.pageViewer.viewport,
            scale: this.pageViewer.scale,
            messageHandler: this.messageHandler,
            pageViewer: this.pageViewer,
            hidden: this.tool === 'addImage' || this.tool === 'editImage'
          })
          this.frameEditorList.push(frameEditor)
          continue
        }

        const isImagePtr = await this.messageHandler.sendWithPromise('IsImageArea', editAreaPtr)
        if (isImagePtr) {
          const frameEditor = new ImageEditor({
            eventBus: this.eventBus,
            contentContainer: this,
            container: this.contentContainer,
            pagePtr: this.pagePtr,
            editPagePtr: this.editPagePtr,
            editAreaPtr,
            editAreaIndex: i,
            viewport: this.pageViewer.viewport,
            scale: this.pageViewer.scale,
            messageHandler: this.messageHandler,
            pageViewer: this.pageViewer,
            $t: this.$t,
            hidden: this.tool === 'addText' || this.tool === 'editText'
          })
          this.frameEditorList.push(frameEditor)
        }
      }
    } finally {
      this.isRendering = false
    }

    if (this.tool === 'addText') {
      this.addTextManager()
    }
    if (this.tool === 'addImage') {
      this.addImageManager()
    }

    this.destroyed = false
    this.rendered = true
  }

  show () {
    this.contentContainer.style.display = 'block'
  }

  hide () {
    this.contentContainer.style.display = 'none'
    this.resetTools()
  }

  cancel () {
    if (this.destroyed || !this.rendered) return

    this.contentContainer?.remove()
    this.contentContainer = null

    this.resetTools()
  }

  destroy () {
    if (this.destroyed || !this.rendered) return

    this.contentContainer?.remove()
    this.contentContainer = null

    this.resetTools()

    this.destroyed = true

    this.eventBus._off('toolChanged', this.onHandleTool)
    document.removeEventListener('keydown', this.onKeydown)

    this.messageHandler.sendWithPromise('ClearEditPage', this.editPagePtr)
  }

  handleTool({
    tool,
    color
  }) {
    this.tool = tool
    this.color = color

    if (this.textManager) {
      this.textManager.tool = tool
      this.textManager.color = color
    }
    if (this.imageManager) {
      this.imageManager.tool = tool
      this.imageManager.color = color
    }

    if (tool === 'addText') {
      if (this.imageManager) {
        this.imageManager.tool = ''
        this.imageManager.reset()
      }
      this.addTextManager()
    }
    if (tool === 'addImage') {
      if (this.textManager) {
        this.textManager.tool = ''
        this.textManager.reset()
      }
      this.addImageManager()
    }

    if (this.tool === 'addText' || this.tool === 'editText') {
      this.eventBus.dispatch('showContentEditorType', 'text')
    } else if (this.tool === 'addImage' || this.tool === 'editImage') {
      this.eventBus.dispatch('showContentEditorType', 'image')
    } else {
      this.eventBus.dispatch('showContentEditorType', '')
    }
  }

  addTextManager () {
    if (this.textManager) {
      this.textManager.init()
    } else {
      this.textManager = new AddText({
        tool: this.tool,
        color: this.color,
        container: this.pageDiv,
        pageViewer: this.pageViewer,
        viewport: this.pageViewer.viewport,
        scale: this.pageViewer.scale,
        page: this.page,
        eventBus: this.eventBus,
        contentContainer: this
      })
    }
  }

  async addImageManager (imageData) {
    if (this.imageManager) {
      this.imageManager.init()
    } else {
      this.imageManager = new AddImage({
        tool: this.tool,
        color: this.color,
        container: this.pageDiv,
        pageViewer: this.pageViewer,
        viewport: this.pageViewer.viewport,
        scale: this.pageViewer.scale,
        page: this.page,
        eventBus: this.eventBus,
        contentContainer: this,
        $t: this.$t
      })
    }
    if (imageData) {
      const { imageBase64, width, height, point, preview } = imageData
      const rect = this.getRect(width, height, point)

      if (preview)  {
        await this.addImageEditor({
          rect,
          imageBase64
        })
        return
      }
      await this.messageHandler.sendWithPromise('CreateNewImageAreaByStream', {
        pagePtr: this.pagePtr,
        editPagePtr: this.editPagePtr,
        rect,
        imageBase64
      })
      await this.messageHandler.sendWithPromise('EndEdit', this.editPagePtr)
    } else {
      this.imageManager.init()
    }
  }

  getRect (width, height, point) {
    const pageWidth = this.viewport.viewBox[2]
    const pageHeight = this.viewport.viewBox[3]
    const fixedWidth = pageWidth / 2
    const fixedHeight = pageHeight / 2

    let scaledWidth = width
    let scaledHeight = height

    if (width > fixedWidth || height > fixedHeight) {
      const imageRatio = width / height
      const rectRatio = fixedWidth / fixedHeight

      if (imageRatio > rectRatio) {
        scaledWidth = fixedWidth
        scaledHeight = fixedWidth / imageRatio
      } else {
        scaledWidth = fixedHeight * imageRatio
        scaledHeight = fixedHeight
      }
    }

    const rect = {
      left: point.x - scaledWidth / 2,
      top: point.y - scaledHeight / 2,
      right: point.x + scaledWidth / 2,
      bottom: point.y + scaledHeight / 2,
    }

    if (rect.left < 7) {
      const offset = 7 - rect.left
      rect.left += offset
      rect.right += offset
    }
    if (rect.top < 7) {
      const offset = 7 - rect.top
      rect.top += offset
      rect.bottom += offset
    }
    if (rect.right > pageWidth) {
      const offset = rect.right - pageWidth + 7
      rect.right -= offset
      rect.left -= offset
    }
    if (rect.bottom > pageHeight) {
      const offset = rect.bottom - pageHeight + 7
      rect.bottom -= offset
      rect.top -= offset
    }
    return rect
  }

  // 添加编辑区域
  async addTextEditor (data) {
    const { rect, fontData, alignType } = data
    const editAreaPtr = await this.messageHandler.sendWithPromise('CreateNewTextArea', {
      pagePtr: this.pagePtr,
      editPagePtr: this.editPagePtr,
      rect,
      fontData,
      alignType
    })

    if (!editAreaPtr) return

    const frameEditor = new TextEditor({
      eventBus: this.eventBus,
      contentContainer: this,
      container: this.contentContainer,
      pagePtr: this.pagePtr,
      editPagePtr: this.editPagePtr,
      editAreaPtr,
      editAreaIndex: this.frameEditorList.length,
      viewport: this.pageViewer.viewport,
      scale: this.pageViewer.scale,
      messageHandler: this.messageHandler,
      newAdd: true,
      pageViewer: this.pageViewer
    })
    this.frameEditorList.push(frameEditor)
  }

  // 添加图片区域
  async addImageEditor (data) {
    const { rect, imageBase64 } = data
    const editAreaPtr = await this.messageHandler.sendWithPromise('CreateNewImageAreaByStream', {
      pagePtr: this.pagePtr,
      editPagePtr: this.editPagePtr,
      rect,
      imageBase64
    })

    if (!editAreaPtr) return

    const frameEditor = new ImageEditor({
      eventBus: this.eventBus,
      contentContainer: this,
      container: this.contentContainer,
      pagePtr: this.pagePtr,
      editPagePtr: this.editPagePtr,
      editAreaPtr,
      editAreaIndex: this.frameEditorList.length,
      viewport: this.pageViewer.viewport,
      scale: this.pageViewer.scale,
      messageHandler: this.messageHandler,
      newAdd: true,
      pageViewer: this.pageViewer,
      $t: this.$t
    })
    this.frameEditorList.push(frameEditor)
  }

  get selectedFrameIndex () {
    return this._selectedFrameIndex
  }

  set selectedFrameIndex (index) {
    this._selectedFrameIndex = index
  }

  removeEditor (index) {
    this.frameEditorList.splice(index, 1)
    this.frameEditorList.forEach((editor, i) => editor.editAreaIndex = i)
  }

  resetTools () {
    if (this.textManager) {
      this.textManager.tool = ''
      this.textManager.reset()
    }
    if (this.imageManager) {
      this.imageManager.tool = ''
      this.imageManager.reset()
    }
  }

  handleKeyDown (e) {
    const selectedEditor = this.frameEditorList[this._selectedFrameIndex]
    if (!selectedEditor || selectedEditor.state !== 1) return

    const keyCode = e.keyCode || e.which
    if (keyCode === 8 || keyCode === 46) { // 8 delete键，46 backspace键
      selectedEditor.remove()
    }
  }

  // undo/redo
  async handleResetOperate(op) {
    if (op === 'undo') {
      const res = await this.historyManager.undo()
      // await this.updateFrameEditor(res)

    } else if (op === 'redo') {
      const res = await this.historyManager.redo()
      // await this.updateFrameEditor(res)
    }
  }

  // 修改后添加到undo列表
  async handleOperateList(data) {
    const canUndo = await this.historyManager.canUndo(data)
    if (!canUndo) return
    this.eventBus.dispatch('undoRedoStatusChanged')
    this.historyManager.undoList.push(data)
    this.historyManager.redoList.length = 0

    this.eventBus.dispatch('changeOperateList', {
      undoListLength: this.historyManager.undoList.length,
      redoListLength: this.historyManager.redoList.length
    })
    this.eventBus.dispatch('isContentEditModyfied', true)
  }

  // undo/redo后，更新视图
  async updateFrameEditor (res) {
    const { editAreaPtr, editCharPlace } = res

    const editor = this.frameEditorList.find(item => item.editAreaPtr === editAreaPtr)
    if (editor) {
      editor.updateCanvasAfterUndoRedo()
      editor.type === 'text' && (editCharPlace.CharIndex !== -1 || editCharPlace.LineIndex !== -1 || editCharPlace.RunIndex !== -1 || editCharPlace.SectionIndex !== -1 ) && editor.updateCursorLine(editCharPlace)
      return
    }

    const newFrameEditorList = []
    const editAreaCount = await this.messageHandler.sendWithPromise('GetEditAreaCount', this.editPagePtr)

    for (let i = 0; i < editAreaCount; i++) {
      const editAreaPtr = await this.messageHandler.sendWithPromise('GetEditArea', {
        editPagePtr: this.editPagePtr,
        index: i
      })
      if (!editAreaPtr) continue

      let frameEditor
      if (await this.messageHandler.sendWithPromise('IsTextArea', editAreaPtr)) {
        frameEditor = new TextEditor({
          eventBus: this.eventBus,
          contentContainer: this,
          container: this.contentContainer,
          pagePtr: this.pagePtr,
          editPagePtr: this.editPagePtr,
          editAreaPtr,
          editAreaIndex: i,
          viewport: this.pageViewer.viewport,
          scale: this.pageViewer.scale,
          messageHandler: this.messageHandler,
          pageViewer: this.pageViewer,
          hidden: this.tool === 'addImage',
          isUpdate: true
        })
      } else if (await this.messageHandler.sendWithPromise('IsImageArea', editAreaPtr)) {
        frameEditor = new ImageEditor({
          eventBus: this.eventBus,
          contentContainer: this,
          container: this.contentContainer,
          pagePtr: this.pagePtr,
          editPagePtr: this.editPagePtr,
          editAreaPtr,
          editAreaIndex: i,
          viewport: this.pageViewer.viewport,
          scale: this.pageViewer.scale,
          messageHandler: this.messageHandler,
          pageViewer: this.pageViewer,
          $t: this.$t,
          hidden: this.tool === 'addText',
          isUpdate: true
        })
      }
      frameEditor && newFrameEditorList.push(frameEditor)
    }

    this.frameEditorList.forEach(async editor => {
      await editor.updateCanvas(null, true)
      editor.removeViewer()
    })
    this.frameEditorList.length = 0
    this.frameEditorList = newFrameEditorList
  }
}
